# *
# *	Copyright (c) 2000 Alberto Reggiori / <alberto.reggiori@jrc.it>
# *	ISIS/RIT, Joint Research Center Ispra (I)
# *
# * NOTICE
# *
# * This product is distributed under a BSD/ASF like license as described in the 'LICENSE'
# * file you should have received together with this source code. If you did not get a
# * a copy of such a license agreement you can pick up one at:
# *
# *     http://xml.jrc.it/RDFStore/LICENSE
# *

package XML::Parser::PenRDFStrawman;
{
	use strict;

	use vars qw{$VERSION %Built_In_Styles $ua};
	use Carp qw(carp croak cluck confess);
	use URI;
	use URI::file;
	use URI::Escape;
	require "chat2.pl";

BEGIN
{
	require XML::Parser::Expat;
    	$VERSION = '0.1';
    	croak "XML::Parser::Expat.pm version 2 or higher is needed"
		unless $XML::Parser::Expat::VERSION =~ /^2\./;
}

$XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
$XML::Parser::PenRDFStrawman::RDF_SCHEMA_NS="http://www.w3.org/TR/1999/PR-rdf-schema-19990303#";
$XML::Parser::PenRDFStrawman::XMLSCHEMA="xml";
$XML::Parser::PenRDFStrawman::XMLNS="xmlns";
$XML::Parser::PenRDFStrawman::RDFMS_parseType = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "parseType";
$XML::Parser::PenRDFStrawman::RDFMS_type = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "type";
$XML::Parser::PenRDFStrawman::RDFMS_about = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "about";
$XML::Parser::PenRDFStrawman::RDFMS_bagID = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "bagID";
$XML::Parser::PenRDFStrawman::RDFMS_resource = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "resource";
$XML::Parser::PenRDFStrawman::RDFMS_aboutEach = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "aboutEach";
$XML::Parser::PenRDFStrawman::RDFMS_aboutEachPrefix = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "aboutEachPrefix";
$XML::Parser::PenRDFStrawman::RDFMS_ID = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "ID";
$XML::Parser::PenRDFStrawman::RDFMS_RDF = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "RDF";
$XML::Parser::PenRDFStrawman::RDFMS_Description = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "Description";
$XML::Parser::PenRDFStrawman::RDFMS_Seq = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "Seq";
$XML::Parser::PenRDFStrawman::RDFMS_Alt = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "Alt";
$XML::Parser::PenRDFStrawman::RDFMS_Bag = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "Bag";
$XML::Parser::PenRDFStrawman::RDFMS_predicate = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "predicate";
$XML::Parser::PenRDFStrawman::RDFMS_subject = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "subject";
$XML::Parser::PenRDFStrawman::RDFMS_object = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "object";
$XML::Parser::PenRDFStrawman::RDFMS_Statement = $XML::Parser::PenRDFStrawman::RDF_SYNTAX_NS . "Statement";

sub new
{
    	my ($class, %args) = @_;
    	my $style = $args{Style};

	$args{_HNDL_TYPES} = {%XML::Parser::Expat::Handler_Setters};
	$args{_HNDL_TYPES}->{Init} = 1;
	$args{_HNDL_TYPES}->{Assert} = 1;
	$args{_HNDL_TYPES}->{Final} = 1;

    	my $handlers = $args{Handlers};
    	if (defined($style))
	{
		my $stylepkg = $style;
		if ($stylepkg !~ /::/)
		{
	    		$stylepkg = "\u$style";
	    		croak "Undefined style: $style" 
				unless defined($Built_In_Styles{$stylepkg});
	    		$stylepkg = 'XML::Parser::PenRDFStrawman::' . $stylepkg;
		}
		unless (defined($handlers))
		{
			$args{Handlers} = $handlers = {};
		}

		my $htype;
		foreach $htype (keys %{$args{_HNDL_TYPES}})
		{
	    		# Handlers explicity given override
	    		# handlers from the Style package
	    		unless (defined($handlers->{$htype}))
			{
				# A handler in the style package must either have
				# exactly the right case as the type name or a
				# completely lower case version of it.
				my $hname = "${stylepkg}::$htype";
				if (defined(&$hname))
				{
		    			$handlers->{$htype} = \&$hname;
		    			next;
				}
				$hname = "${stylepkg}::\L$htype";
				if (defined(&$hname))
				{
		    			$handlers->{$htype} = \&$hname;
		    			next;
				}
	    		}
		}
	}
    	$args{Pkg} ||= caller;

    	bless \%args, $class;
}

sub setHandlers
{
	my ($class, @handler_pairs) = @_;

	croak("Uneven number of arguments to setHandlers method") 
		if (int(@handler_pairs) & 1);

	my @ret;
        while (@handler_pairs)
        {
                my $type = shift @handler_pairs;
                my $handler = shift @handler_pairs;
                unless (defined($class->{_HNDL_TYPES}->{$type})) {
                        my @types = sort keys %{$class->{_HNDL_TYPES}};
                        croak("Unknown Parser handler type: $type\n Valid types are : @types");
                }
                push(@ret, $type, $class->{Handlers}->{$type});
                $class->{Handlers}->{$type} = $handler;
        };
        return @ret;
};

sub setSource {
  	my ($file_or_uri)=@_;
	$file_or_uri .= '#'
		if(	(defined $file_or_uri) &&
			(!($file_or_uri =~ /#$/)) ||
			($file_or_uri =~ /\/$/) ||
			($file_or_uri =~ /:$/)
			);
	return $file_or_uri;
};

sub parse
{
	my ($class,$arg,$file_or_uri) = @_;

	my @expat_options = ( Namespaces => 1 ); #RDF needs them :)
	my ($key, $val);
	while (($key, $val) = each %{$class})
      	{
		push(@expat_options, $key, $val) 
			unless exists $class->{Non_Expat_Options}->{$key};
      	}

	#Run Expat
    	my $first = new XML::Parser::Expat(@expat_options, @_);

	$first->{PenRDFStrawman}={};

	croak "Missing NodeFactory"
		unless(defined $class->{NodeFactory});
  	$first->{PenRDFStrawman}->{nodeFactory} = $class->{NodeFactory};

    	$first->setHandlers(	Start => \&RDFXML_StartElementHandler,
				End => \&RDFXML_EndElementHandler,
				Char => \&RDFXML_CharacterDataHandler );
	$first->{Extra_Options} = $class->{Extra_Options};

	$class->{handlers} = {};
    	%{$class->{handlers}} = %{$class->{Handlers}} 
		if(defined $class->{Handlers});

    	my $init = delete $class->{handlers}->{Init};
    	my $final = delete $class->{handlers}->{Final};

	#Trigger 'Init' event
    	&$init($first) 
		if defined($init);

    	my $result = $first->parse($arg);

	$first->{parser_parameters} = [];
	@{$first->{parser_parameters}} = (@expat_options,@_);

    	if (defined($final))
	{
		#Trigger 'Final' event
		$result = &$final($first);
	}

    	return $result;
}

sub parsestring
{
	my $class = shift;
	return $class->parse(@_);
}

sub parsefile
{
	my $class = shift;
	my $file = shift;

	if( (defined $file) && ($file ne '') ) {
		my $ret;
		my $file_uri;
		if(-e $file) {
                        $file_uri= URI::file->new_abs($file);
                } elsif(!($file =~ /^[\w]+:/)) {
                        $file_uri= URI::file->new_abs($file,URI::file->cwd);
                } else {
                        $file_uri= URI->new($file);
                };	
		if (	(defined $file_uri->scheme)	&&
			($file_uri->scheme ne 'file') ) {
    			my $uri = new URI($file);
			my $wget_handle = wget($uri);
                        if(defined $wget_handle) {
                                eval {
                                        $ret = $class->parse($wget_handle, $uri);
                                };
                                my $err = $@;
                                croak $err
                                        if $err;
                        } else {
                                croak "Cannot fetch '$uri'";
                        };
    		} else {
			my $file= $file_uri->file;

			croak "Unable to open input RDF file ($file)!"
				unless -e $file;
			open(FF, $file) 

				or  croak "Couldn't open $file:\n$!";
			eval {
				$ret = $class->parse(*FF, $file_uri);
    			};
    			my $err = $@;
    			close(FF);
    			croak $err 	
				if $err;
		};
    		return $ret;
  	};
}

sub wget {
        my ($uri) = @_;

        croak "wget: input url is not an instance of URI"
                unless( (defined $uri) && ($uri->isa("URI")) );

        my $handle = &chat::open_port($uri->host,$uri->port);
        &chat::print($handle,"GET $uri HTTP/1.0\r\nUser-Agent: Alberto.Reggiori\@jrc.it\r\n\r\n");
        *S = *chat::S;
        my $line = <S>;
        if (!($line =~ m#^HTTP/(\d+)\.(\d+) (\d\d\d) (.+)$#)) {
                &chat::close($handle);
                warn "Did not get HTTP/X.X header back...$line";
                return undef;
        };
        my $status = $3;
        my $reason = $4;
        if ( $status != 200 ) {
                &chat::close($handle);
                warn "Error MSG returned from server: $status $reason\n";
                return undef;
        };
        while(<S>) {
                chomp;
                last if m/^\s+$/;
        };
        return *S;
};

sub getAttributeValue {
	my ($expat,$attlist, $elName) = @_;

  	return undef
		if(!@{$attlist});

	my $n;
	for($n=0; $n<=$#{$attlist}; $n+=2) {
    		my $attname = $attlist->[$n];
    		return $attlist->[$n+1]
			if ($attname eq $elName);
  	};
  	return undef;
}

sub RDFXML_StartElementHandler {
	my $expat = shift;
	my $tag = shift;
	my @attlist = @_;

	my $sNamespace = $expat->namespace($tag);
	if(not(defined $sNamespace)) {			
		my ($prefix,$suffix) = split(':',$tag);
		if($prefix eq $XML::Parser::PenRDFStrawman::XMLSCHEMA) {
			$sNamespace = $XML::Parser::PenRDFStrawman::XMLSCHEMA;
			$tag = $sNamespace.$suffix; #xml:lang ---> xmllang
		} else {
			$expat->xpcroak("Unresolved namespace prefix ",$prefix);
		};
	} else {
		$tag = $sNamespace.$tag;
	};
	my $n;
	for($n=0; $n<=$#attlist; $n+=2) {
    		my $attname = $attlist[$n];
		my $namespace = $expat->namespace($attname);
		unless(defined $namespace) { #default namespace
			my ($prefix,$suffix) = split(':',$attname);
			if( (defined $prefix) && (defined $suffix) ) {
				if($prefix eq $XML::Parser::PenRDFStrawman::XMLSCHEMA) {
					$namespace = $XML::Parser::PenRDFStrawman::XMLSCHEMA;
					$attlist[$n] = $namespace.$suffix;
				} else {
					$expat->xpcroak("Unresolved namespace prefix ",$prefix);
				};
			} else {
				$namespace = $sNamespace;
				$attlist[$n]=$namespace.$attname;
			};
		} else {
			$attlist[$n] = $namespace.$attname;
		};
  	};

	if( $tag eq $XML::Parser::PenRDFStrawman::RDFMS_subject ) {
		$expat->{RDF_subject} = $expat->{PenRDFStrawman}->{nodeFactory}->createResource(getAttributeValue($expat,\@attlist,$XML::Parser::PenRDFStrawman::RDFMS_resource));
        } elsif( $tag eq $XML::Parser::PenRDFStrawman::RDFMS_predicate ) {
		$expat->{RDF_predicate} = $expat->{PenRDFStrawman}->{nodeFactory}->createResource(getAttributeValue($expat,\@attlist,$XML::Parser::PenRDFStrawman::RDFMS_resource));
        } elsif( $tag eq $XML::Parser::PenRDFStrawman::RDFMS_object ) {
		if(my $att=getAttributeValue($expat,\@attlist,$XML::Parser::PenRDFStrawman::RDFMS_resource)) {
			#create RDFNode if necessary
			$expat->{RDF_object} = $expat->{PenRDFStrawman}->{nodeFactory}->createResource($att);
		} else {
			#reset Literal
			$expat->{RDF_literal}='';
		};
	};
}

sub RDFXML_EndElementHandler {
	my $expat = shift;
	my $tag = shift;

	my $sNamespace = $expat->namespace($tag);
	if(not(defined $sNamespace)) {			
		my ($prefix,$suffix) = split(':',$tag);
		if($prefix eq $XML::Parser::PenRDFStrawman::XMLSCHEMA) {
			$sNamespace = $XML::Parser::PenRDFStrawman::XMLSCHEMA;
			$tag = $sNamespace.$suffix; #xml:lang ---> xmllang
		} else {
			$expat->xpcroak("Unresolved namespace prefix ",$prefix);
		};
	} else {
		$tag = $sNamespace.$tag;
	};


	if( $tag eq $XML::Parser::PenRDFStrawman::RDFMS_Statement ) {
		#create Statement
		my $st = $expat->{PenRDFStrawman}->{nodeFactory}->createStatement($expat->{RDF_subject},$expat->{RDF_predicate},$expat->{RDF_object});

		#Trigger 'Assert' event
        	my $assert = $expat->{Handlers}->{Assert};
        	if (defined($assert)) {
        		&$assert($expat, $st);
		};

		#wipe up
		delete $expat->{RDF_subject};
		delete $expat->{RDF_predicate};
		delete $expat->{RDF_object};
		delete $expat->{RDF_literal};
        } elsif(	($tag eq $XML::Parser::PenRDFStrawman::RDFMS_object ) &&
			(!(defined $expat->{RDF_object})) ) {
		#create Literal if necessary
		$expat->{RDF_object} = $expat->{PenRDFStrawman}->{nodeFactory}->createLiteral($expat->{RDF_literal});
	};
};

sub RDFXML_CharacterDataHandler {
	my $expat = shift;
	my $text = shift;

	if(defined $expat->{RDF_literal}) { #consider just object data
     		# Warning: this is not correct procedure according to XML spec.
		# All whitespace matters!
		# skip preserveWhiteSpace() by checking here....
		#	return
		#		if( ($text =~ /^\s+/) && ($text =~ /\s+$/) && (!$bHasData) );
		# trim text
		my $trimtext = $text;
		$trimtext =~ s/^([ ])+//g;
		$trimtext =~ s/([ ])+$//g;
		if(length($trimtext)>0) {
			#add data to object Literal
			$expat->{RDF_literal} .= $text;
		};
	};
};

1;
};

__END__

=head1 NAME

XML::Parser::PenRDFStrawman - This module implements an RDF strawman parser for the syntax proposed by Jonathan Borden at http://www.openhealth.org/RDF/rdf_Syntax_and_Names.htm

=head1 SYNOPSIS

	use XML::Parser::PenRDFStrawman;
        use RDF::Store::NodeFactory;
        my $p=new XML::Parser::PenRDFStrawman(
		ErrorContext => 2,
                Handlers        => {
                        Init    => sub { print "INIT\n"; },
                        Final   => sub { print "FINAL\n"; },
                        Assert  => sub { print "STATEMENT - @_\n"; }
                },
                NodeFactory     => new RDF::Store::NodeFactory() );

	$p->parsefile('http://www.openhealth.org/RDF/mns-fig13-result.xml');
        $p->parsefile('/some/where/my.xml');
        $p->parsefile('file:/some/where/my.xml');
	$p->parse(*STDIN);

	use XML::Parser::PenRDF::RDFStore;
	my $pstore=new XML::Parser::PenRDFStrawman(
                ErrorContext 	=> 2,
                Style 		=> 'XML::Parser::PenRDF::RDFStore',
                NodeFactory     => new RDF::Store::NodeFactory(),
                store   =>      {
                                	persistent      =>      1,
                                	directory       =>      '/tmp/',
                                	seevalues       =>      1,
                                	options         =>      { style => 'BerkeleyDB', Q => 20 }
                                }
        );
	$pstore->parsefile('http://www.openhealth.org/RDF/mns-sect7.2-result.xml');

=head1 DESCRIPTION

This module implements a Resource Description Framework (RDF) strawman parser compliant to the syntax proposed by Jonathan Borden at http://www.openhealth.org/RDF/rdf_Syntax_and_Names.htm using the XSLT style sheet at http://www.openhealth.org/RDF/extract/rdfExtractity.xsl. Such a syntax is yet another extension/refinement of the original syntax proposed by Dan Connoly at http://www.w3.org/XML/2000/04rdf-parse/rdfp.xsl and already extended by Jason Diamond's at http://www.injektilo.org/rdf/rdf.xsl. The parser has been completely written in Perl using the XML::Parser::Expat(3) module. For the actual explaination see the XML::Parser::PenRDF(3) man page.

=head1 METHODS

XML::Parser::PenRDFStrawman supports all the XML::Parser::PenRDF options but Source. See the manual page for XML::Parser::PenRDF(3)

=head1 BUGS

Although the syntax proposed by Jonathan Borden is quite complete, the style-sheet is not unfortunately compatible with the Sablotron(3) great XSLT engine, that does not support 'exclude-result-prefixes' and functions overriding. A modified version working for Sablotron is available in the samples directory in the file xml2rdf.xsl.

=head1 SEE ALSO

XML::Parser::PenRDF(3), Sablotron(3) RDF::Store::NodeFactory(3)

=item

Mapping namespace qualitied element names to URIs - http://www.openhealth.org/RDF/QNameToURI.htm

=item

Jason Diamond's http://injektilo.org/rdf/examples.html

=item

Dan Connoly strawman sytax - http://www.w3.org/XML/2000/04rdf-parse/

=item

TimBL's semantic web toolbox - http://www.w3.org/DesignIssues/Toolbox

=item

RDF Model and Syntax Specification - http://www.w3.org/TR/REC-rdf-syntax

=item

RDF Schema Specification 1.0 - http://www.w3.org/TR/2000/CR-rdf-schema-20000327

=head1 AUTHOR

	Alberto Reggiori <alberto.reggiori@jrc.it>
	Clark Cooper is the author of the XML::Parser(3) module together with Larry wall
